/*
 * file proj3.c:  symbol table data structures and routines. Xiaolin Zang,
 * fall 1989 for cs 221, project 3 
 *
 * Modified Oct 91 for CS2211, C-N. Fiechter
 *
 */

/*
 * TO USE THIS MODULE YOU HAVE TO DEFINE:
 * =====================================
 *
 * FILE *table   \
 * getstring()    }  as for proj2.c
 * getname()     / 
 *
 * 
 * void 
 * error_msg(error_type, proc_instr, id);
 *   int error_type, proc_instr, id;
 * {
 *    Report an error, the type of which is given by 'error_type'
 *    (see the defined error_type below). 
 *
 *    If 'id' is not zero it is the unique identifying number of the 
 *    identifier for which the error as to be reported.
 *
 *    Proceed according to 'proc_instr' (ABORT / CONTINUE, see below in
 *    the define section).
 * }
 *
 * 
 * AND TO INSERT THE ADEQUATE CODE:
 * ===============================
 *
 * - in STInit() 
 *
 * - in STPrint()  (lines 453 and 457)
 * 
 */


/*
 * there are three arrays for symbol table operation.  "st" is the real
 * symbol table, it carries all the information of an id and will still be
 * used in code generation phase.  The attributes for an id are kept in a
 * link list pointed by its symbol table entry. It is easy to be extended in
 * the next phase to include more attributes, such like the size and the
 * location of a varialbe. Below are some sample attributes of an id.  You
 * should design your own. "stack" is a temporary structure in which all the
 * id's in the current scoping context are visible.  "attrarray" is to store
 * all the attributes in symbol table. 
 */

#define false 0
#define true 1
#define FALSE 0
#define TRUE 1
#define bool char

#define STACK_SIZE 100
#define ST_SIZE 500
#define ATTR_SIZE 2000

/*
 * error type for error reporting routine 
 */
#define STACK_OVERFLOW 100
#define REDECLARATION 101
#define ST_OVERFLOW 102
#define UNDECLARATION 103
#define ATTR_OVERFLOW 104
#define NOT_USED 105

/*
 * processing instruction for error reporting routine 
 */
#define CONTINUE 0	     /* print error and return to the caller */
#define ABORT 1		     /* print the fatal error and abort execution */

/*
 * these definitions correspond to the fields of a stack element 
 */
#define MARKER 1
#define NAME 2
#define ST_PTR 3
#define DUMMY 4
#define USED 5

/*
 * the possible attributes for symbol table.  the comment to the right
 * describe the attribute's value.  Notice the small constants are given to
 * the attributes which are common to all the ids, so that we can do some
 * sorting in the link list 
 */
#define NAME_ATTR 1	     /* value: id lexeme pointer, set by InsertEntry */
#define NEST_ATTR 2	     /* value: nesting level, set by InsertEntry */
#define TREE_ATTR 3	     /* value: point back to the subtree */
#define PREDE_ATTR 4	     /* value: is this id predefined? */
#define TYPE_ATTR 6	     /* value: pointer to the type tree for a
			      * varible, constant id or function */
#define VALUE_ATTR 7	     /* value: the value of a constant id (integer,
			      * charater or string pointer) */
#define KIND_ATTR 5	     /* value: see below */

/*
 * the possible values of attribute kind_attr 
 */
#define CONST 1
#define VAR 2
#define PROCE 3
#define FUNC 4
#define VAR_ARG 5
#define VALUE_ARG 6
#define FIELD 7

/*********************************data structures**********************/

/*
 * !!!!!!!!! WARNING !!!!!!!!!! all the data structures and variables defined
 * below are private to the symbol table operating routines. Do NOT use them
 * directly unless you really understant the whole symbol table structure. 
 */

/*
 * struct of symbol table stack 
 */
struct
{
  bool marker;		     /* mark the beginning of a block */
  int name;		     /* point to the lexeme of the id */
  int st_ptr;		     /* point to the id's symble table entry */
  bool dummy;		     /* dummy element to indicate a undeclared id */
  bool used;		     /* this id is used? */
} stack[STACK_SIZE];	     /* stack array */

/*
 * struct of element of attribute list 
 */
typedef struct
{
  char attr_num;	     /* attribute number ( < 256 ) */
  int attr_val;		     /* attribute value */
  int next_attr;
} attr_type;		     /* define the struct type and the pointer type */

/*
 * symbol table array 
 */
int st[ST_SIZE];

/*
 * attribute array.  attributes are allocated from this array instead of from
 * dynamically allocated space.  It may easy testing and debugging. 
 */
attr_type attrarray[ATTR_SIZE];

/*
 * global counters 
 */
int stack_top = 0;	     /* stack top counter */
int st_top = 0;		     /* symbol table top counter */
int nesting = 0;	     /* nesting level counter */
int attr_top = 0;	     /* attribute array counter */

/************************ routines *****************************/

extern void Push();
extern void SetAttr();
extern void error_msg();

/*
 * STInit():  initialize the symbol table.  put predefined names into it and
 * build trees for those names. this routine is incomplete, please fill in
 * the details 
 */
void
STInit()
{
  /*
   * initialze the predefine variables, types, and functions
   * (!!! your work !!!) 
   * Example for insert the integer type
   * index = hash_insert("integer");
   * STIndex = InsertEntry(index);
   * SetAttr(STIndex, TYPE_ATTR,(int)IntExp());
   * SetAttr(STIndex, PREDE_ATTR, true);
   * SetAttr(STIndex, KIND_ATTR, TYPE);
   */

  /*
   * build syntax tree for function chr and put it into symbol table 
   * (!!! your work !!!) 
   */

  /*
   * put parameter i of function chr into symbol table
   * (!!! your work !!!) 
   */


  /*
   * build syntax tree and symbol table entry for function ord 
   * (!!! your work !!!) 
   */

  /*
   * build syntax tree for procedure write and put it into symbol table 
   * (!!! your work !!!) 
   */
 

  /* and so on ..... */


}


/*
 * InsertEntry():  builds a symbol table entry for id.  the current block is
 * searched for redeclaration error.  the id's name  and nesting level
 * attributes are set by this function.  writing error_msg routine is your
 * work. 
 */
int
InsertEntry(id)
  int id;
{
  /* id is already declared in the current block */
  if (LookUpHere(id))
  {
    error_msg(REDECLARATION, CONTINUE, id);
    return (0);
  }

  if (st_top >= ST_SIZE-1)
    error_msg(ST_OVERFLOW, ABORT);

  st_top++;
  st[st_top] = 0;
  SetAttr(st_top, NAME_ATTR, id);
  SetAttr(st_top, NEST_ATTR, nesting);
  Push(false, id, st_top, false);
  return (st_top);
}

/*
 * LookUp(): search an id in the stack and return the symbol table entry
 * pointer, if found.  if the id is not in the stack, report error and push a
 * dummy item on the stack so that the same error will not be reported
 * repeatedly 
 */
int
LookUp(id)
  int id;
{
  int i;

  for (i = stack_top; i > 0; i--)
    if (!stack[i].marker && stack[i].name == id)
    {
      stack[i].used = true;
      return (stack[i].st_ptr);
    }

  /* id is undefined, push a dummy element onto stack */
  error_msg(UNDECLARATION, CONTINUE, id);
  Push(false, id, 0, true);
  return 0;
}

/*
 * LookUpHere():  search an id in the stack for the current block.  It
 * returns the symbol table pointer if the id is found.  otherwise, return 0
 * this routine can be used to check if there is a forward declaration
 * for a procedure/function
 */
int
LookUpHere(id)
  int id;
{
  int i;

  for (i = stack_top; !stack[i].marker; i--)
    if (stack[i].name == id && !stack[i].dummy)
      return (stack[i].st_ptr);
  return (0);
}

/*
 * LookUpField():  it is found unnecessary.  so it is not implemented 
 */

/*
 * OpenBlock():  build a mark on the stack to indecate the beginning of a new
 * block. increment nesting level counter.  It is called when reserved words
 * "program", "procedure", "function" or "record" is encountered.  Note:
 * procedure or function name should be inserted into symbol table before
 * calling this routine. 
 */
void
OpenBlock()
{
  nesting++;
  Push(true, 0, 0, false);		 /* push a marker onto stack */
}

/*
 * CloseBlock():  decrement nesting level counter and remove the current
 * block from the stack.  called when exitting from a program, procedure,
 * function or a record definition. each element is checked to see if it is
 * used in the block 
 */
void
CloseBlock()
{
  int i;

  for (i = stack_top; !stack[i].marker; i--)
    if (!stack[i].used && !stack[i].dummy)
      error_msg(NOT_USED, CONTINUE, stack[i].name);

  nesting--;
  stack_top = i - 1;				 /* trim the stack */
}

/*
 * IsAttr(): return the index to the searched attribute in attrarray if it is
 * in (nonzero).  otherwise, return false. 
 */
int
IsAttr(st_ptr, attr_num)
  int st_ptr, attr_num;

{
  int i;

  i = st[st_ptr];
  while (i)
  {
    if (attrarray[i].attr_num == attr_num)
      return (i);
    if (attrarray[i].attr_num > attr_num)
      break;
    i = attrarray[i].next_attr;
  }
  return (0);
}

/*
 * GetAttr(): return the specified attribute value for a symbol table entry
 * if found.  otherwise, report error.  Note, this error is not the fault of
 * source program but of the compiler writer.  It is printed for testing and
 * debugging. 
 */
int
GetAttr(st_ptr, attr_num)
  int st_ptr, attr_num;

{
  int i;

  i = IsAttr(st_ptr, attr_num);
  if (!i)
  {
    printf("DEBUG--The wanted attribute number %d does not exist\n", attr_num);
    return (0);
  }

  return (attrarray[i].attr_val);
}

/*
 * SetAttr(): set attribute.  if the attribute is already there,  print
 * debugging message. attributes for a symbol table entry are sorted by their
 * attr_num.  
 */
void
SetAttr(st_ptr, attr_num, attr_val)
  int st_ptr, attr_num, attr_val;

{
  int *p, next;

  if (IsAttr(st_ptr, attr_num))
  {
    printf("DEBUG--The attribute number %d to be added already exists\n",
            attr_num);
    return;
  }

  p = &st[st_ptr];
  next = st[st_ptr];

  /* search the link list for the right insert position */
  while (next)
  {
    if (attrarray[next].attr_num < attr_num)
    {
      p = &(attrarray[next].next_attr);
      next = attrarray[next].next_attr;
    }
    else
      break;
  }

  if (attr_top >= ATTR_SIZE-1)
    error_msg(ATTR_OVERFLOW, ABORT,0);

  attr_top++;
  attrarray[attr_top].attr_num = attr_num;
  attrarray[attr_top].attr_val = attr_val;
  attrarray[attr_top].next_attr = next;
  *p = attr_top;
}

#include <stdio.h>

extern FILE *table;

char *kind_name[] =
{"constant", "variable", "procedure", "function", "var_arg", "val_arg", "field"
};

/*
 * STPrint():  print symbol table.  This routine is incomplete, please fill
 * in details. 
 */
void
STPrint()
{
  int i, attr_num, attr_val, attr_ptr, type_attr;

  fprintf(table,
	  "********************************Symbol Table************************************\n\n");
  fprintf(table,
	  "          Name Nest-Level  Tree-Node Predefined       Kind       Type      Value\n\n");

  for (i = 1; i <= st_top; i++)
  {
    fprintf(table, "%3d", i);
    for (attr_num = NAME_ATTR; attr_num <= VALUE_ATTR; attr_num++)
      if (attr_ptr = IsAttr(i, attr_num))
      {
	attr_val = attrarray[attr_ptr].attr_val;
	switch (attr_num)
      {
	 case NAME_ATTR:
	  fprintf(table, "%11s", getname(attr_val));
	  break;
	 case NEST_ATTR:
	 case TREE_ATTR:
	 case TYPE_ATTR:
	  fprintf(table, "%11d", attr_val);
	  break;
	 case PREDE_ATTR:
	  fprintf(table, "%11s", "yes");
	  break;
	 case KIND_ATTR:
	  fprintf(table, "%11s", kind_name[attr_val - 1]);
	  break;
	 case VALUE_ATTR:
	  if ((type_attr = GetAttr(i, TYPE_ATTR)) == 
              "pointer to Char Type")  /* replace by adequate code */
	    fprintf(table, "%11c", (char) attr_val);
	  else
	  if (type_attr == 
              "pointer to Int Type")   /* replace as above */
	    fprintf(table, "%11d", attr_val);
	  else
	    fprintf(table, "%11s", getstring(attr_val));
	  break;
      }
      }
      else
	fprintf(table, "%11s", " ");
    fprintf(table, "\n");
  }
}


/*
 * Push():  push an element onto the stack.  
 */
void
Push(marker, name, st_ptr, dummy)
     int marker,name,st_ptr,dummy;
{

  if (stack_top >= STACK_SIZE-1)
    error_msg(STACK_OVERFLOW, ABORT,0);

  stack_top++;

  stack[stack_top].marker = (bool) marker;
  stack[stack_top].name = name;
  stack[stack_top].st_ptr = st_ptr;
  stack[stack_top].dummy = (bool) dummy;
  stack[stack_top].used = false;

}


